// written by Daniel Oaks <daniel@danieloaks.net>
// released under the ISC license

/*
Package ircutils provides small, useful utility functions and classes.

This package is in an alpha stage.
*/
package ircutils
